/* dCache - http://www.dcache.org/
 *
 * Copyright (C) 2021-2022 Deutsches Elektronen-Synchrotron
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.dcache.util.files;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.function.Function;
import java.util.function.Supplier;
import org.dcache.util.files.LineBasedParser.UnrecoverableParsingException;
import org.dcache.util.Result;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class parses the contents of a file.  It does this by assuming the file is written in UTF-8
 * and may be split into separate lines that can be parsed where each line is presented one after
 * the other.
 * <p>
 * The actual parsing of the lines is handled by another object: some instance of LineBasedParser.
 *
 * @param <T> The type of state represented the parsed file.
 */
public class LineByLineParser<T> implements Function<Path, Result<T,String>> {

    private static final Logger LOGGER = LoggerFactory.getLogger(LineByLineParser.class);

    private final Supplier<LineBasedParser<T>> parserFactory;

    public LineByLineParser(Supplier<LineBasedParser<T>> parserFactory) {
        this.parserFactory = parserFactory;
    }

    @Override
    public Result<T,String> apply(Path file) {
        LineBasedParser<T> parser = parserFactory.get();

        int lineNumber = 1;
        try {
            for (String line : Files.readAllLines(file)) {
                parser.accept(line);
                lineNumber++;
            }
        } catch (IOException e) {
            LOGGER.warn("{}: {}", file, e.toString());
            return Result.failure(file + " " + e);
        } catch (UnrecoverableParsingException e) {
            LOGGER.warn("{}:{} {}", file, lineNumber, e.getMessage());
            return Result.failure(file + ":" + lineNumber + " " + e.getMessage());
        }

        T result = parser.build();
        return Result.success(result);
    }
}
