/*
COPYRIGHT STATUS:
Dec 1st 2001, Fermi National Accelerator Laboratory (FNAL) documents and
software are sponsored by the U.S. Department of Energy under Contract No.
DE-AC02-76CH03000. Therefore, the U.S. Government retains a  world-wide
non-exclusive, royalty-free license to publish or reproduce these documents
and software for U.S. Government purposes.  All documents and software
available from this server are protected under the U.S. and Foreign
Copyright Laws, and FNAL reserves all rights.

Distribution of the software available from this server is free of
charge subject to the user following the terms of the Fermitools
Software Legal Information.

Redistribution and/or modification of the software shall be accompanied
by the Fermitools Software Legal Information  (including the copyright
notice).

The user is asked to feed back problems, benefits, and/or suggestions
about the software to the Fermilab Software Providers.

Neither the name of Fermilab, the  URA, nor the names of the contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

DISCLAIMER OF LIABILITY (BSD):

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED  WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED  WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL FERMILAB,
OR THE URA, OR THE U.S. DEPARTMENT of ENERGY, OR CONTRIBUTORS BE LIABLE
FOR  ANY  DIRECT, INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT  OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE  POSSIBILITY OF SUCH DAMAGE.

Liabilities of the Government:

This software is provided by URA, independent from its Prime Contract
with the U.S. Department of Energy. URA is acting independently from
the Government and in its own private capacity and is not acting on
behalf of the U.S. Government, nor as its contractor nor its agent.
Correspondingly, it is understood and agreed that the U.S. Government
has no connection to this software and in no manner whatsoever shall
be liable for nor assume any responsibility or obligation for any claim,
cost, or damages arising out of or resulting from the use of the software
available from this server.

Export Control:

All documents and software available from this server are subject to U.S.
export control laws.  Anyone downloading information from this server is
obligated to secure any necessary Government licenses before exporting
documents or software obtained from this server.
 */
package org.dcache.resilience;

import java.util.Collection;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.FutureTask;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.stream.Collectors;
import org.dcache.resilience.util.ExceptionMessage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class TestSynchronousExecutor implements ScheduledExecutorService {

    public enum Mode {
        NOP, RUN, FAIL
    }

    final Logger logger = LoggerFactory.getLogger(this.getClass());

    final Mode mode;

    public TestSynchronousExecutor(Mode mode) {
        this.mode = mode;
    }

    @Override
    public boolean awaitTermination(long timeout, TimeUnit unit)
          throws InterruptedException {
        return false;
    }

    @Override
    public void execute(Runnable command) {
        switch (mode) {
            case NOP:
            case FAIL:
                break;
            default:
                command.run();
        }
    }

    @Override
    public <T> List<Future<T>> invokeAll(
          Collection<? extends Callable<T>> tasks)
          throws InterruptedException {
        return tasks.stream().map((t) -> submit(t)).collect(
              Collectors.toList());
    }

    @Override
    public <T> List<Future<T>> invokeAll(
          Collection<? extends Callable<T>> tasks, long timeout,
          TimeUnit unit) throws InterruptedException {
        return invokeAll(tasks);
    }

    @Override
    public <T> T invokeAny(Collection<? extends Callable<T>> tasks)
          throws InterruptedException, ExecutionException {
        Callable<T> chosen = tasks.iterator().next();
        submit(chosen);
        return null;
    }

    @Override
    public <T> T invokeAny(Collection<? extends Callable<T>> tasks,
          long timeout, TimeUnit unit)
          throws InterruptedException, ExecutionException,
          TimeoutException {
        return invokeAny(tasks);
    }

    @Override
    public boolean isShutdown() {
        return false;
    }

    @Override
    public boolean isTerminated() {
        return false;
    }

    @Override
    public ScheduledFuture<?> schedule(Runnable command, long delay,
          TimeUnit unit) {

        execute(command);
        return null;
    }

    @Override
    public <V> ScheduledFuture<V> schedule(Callable<V> callable, long delay,
          TimeUnit unit) {
        submit(callable);
        return null;
    }

    @Override
    public ScheduledFuture<?> scheduleAtFixedRate(Runnable command,
          long initialDelay,
          long period, TimeUnit unit) {
        execute(command);
        return null;
    }

    @Override
    public ScheduledFuture<?> scheduleWithFixedDelay(Runnable command,
          long initialDelay,
          long delay,
          TimeUnit unit) {
        execute(command);
        return null;
    }

    @Override
    public void shutdown() {
    }

    @Override
    public List<Runnable> shutdownNow() {
        return null;
    }

    @Override
    public <T> Future<T> submit(Callable<T> task) {
        try {
            switch (mode) {
                case NOP:
                case FAIL:
                    break;
                default:
                    task.call();
            }
        } catch (Exception e) {
            logger.error("{}", new ExceptionMessage(e));
        }
        return new FutureTask<T>(task);
    }

    @Override
    public <T> Future<T> submit(Runnable task, T result) {
        try {
            execute(task);
        } catch (Exception e) {
            logger.error("{}", new ExceptionMessage(e));
        }
        return new FutureTask(task, task);
    }

    @Override
    public Future<?> submit(Runnable task) {
        try {
            execute(task);
        } catch (Exception e) {
            logger.error("{}", new ExceptionMessage(e));
        }
        return new FutureTask(task, task);
    }
}
