/* dCache - http://www.dcache.org/
 *
 * Copyright (C) 2014 Deutsches Elektronen-Synchrotron
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.dcache.xrootd.plugins.tls;

import com.google.common.base.Throwables;
import eu.emi.security.authn.x509.CrlCheckingMode;
import eu.emi.security.authn.x509.NamespaceCheckingMode;
import eu.emi.security.authn.x509.OCSPCheckingMode;
import io.netty.handler.ssl.SslContext;
import java.io.File;
import java.util.Properties;
import java.util.function.Supplier;

import org.dcache.ssl.CanlContextFactory;

/**
 * Provides the SSL handler's context using the CanlContextFactory.
 * <p/>
 * This class's FQN should be assigned to the xrootd property:<br/>
 * <p>
 * xrootd.security.tls.handler-factory.class=org.dcache.xrootd.plugins.tls.CDCCanlTLSHandlerFactory
 */
public class CDCCanlTLSHandlerFactory extends SSLHandlerFactory {

    public static final String SERVICE_KEY = "xrootd.security.tls.hostcert.key";
    public static final String SERVICE_CERT = "xrootd.security.tls.hostcert.cert";
    public static final String SERVICE_CACERTS = "xrootd.security.tls.ca.path";
    public static final String NAMESPACE_MODE = "xrootd.security.tls.ca.namespace-mode";
    public static final String CRL_MODE = "xrootd.security.tls.ca.crl-mode";
    public static final String OCSP_MODE = "xrootd.security.tls.ca.ocsp-mode";

    @Override
    protected Supplier<SslContext> buildContextSupplier(Properties properties) throws Exception {
        File serviceKey = new File(properties.getProperty(SERVICE_KEY));
        File serviceCert = new File(properties.getProperty(SERVICE_CERT));
        File serviceCaCerts = new File(properties.getProperty(SERVICE_CACERTS));
        NamespaceCheckingMode namespaceMode =
              NamespaceCheckingMode.valueOf(properties.getProperty(NAMESPACE_MODE));
        CrlCheckingMode crlMode
              = CrlCheckingMode.valueOf(properties.getProperty(CRL_MODE));
        OCSPCheckingMode ocspMode
              = OCSPCheckingMode.valueOf(properties.getProperty(OCSP_MODE));

        var builder = CanlContextFactory.custom()
              .withCertificatePath(serviceCert.toPath())
              .withKeyPath(serviceKey.toPath())
              .withCertificateAuthorityPath(serviceCaCerts.toPath())
              .withCrlCheckingMode(crlMode)
              .withOcspCheckingMode(ocspMode)
              .withNamespaceMode(namespaceMode)
              .withLazy(false)
              .startTls(startTls)
              .buildWithCaching(SslContext.class);

        return () -> {
            try {
                return builder.call();
            } catch (Exception e) {
                Throwables.throwIfUnchecked(e);
                throw new RuntimeException(e);
            }
        };
    }
}
