/* dCache - http://www.dcache.org/
 *
 * Copyright (C) 2017 Deutsches Elektronen-Synchrotron
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.dcache.macaroons;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkState;
import static java.util.Objects.requireNonNull;
import static org.dcache.macaroons.InvalidCaveatException.checkCaveat;

import com.google.common.io.BaseEncoding;
import diskCacheV111.util.FsPath;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.Arrays;
import java.util.Base64;
import java.util.EnumSet;
import java.util.Optional;
import java.util.OptionalLong;
import java.util.stream.LongStream;
import org.dcache.auth.attributes.Activity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class holds information that is (or is to be) encoded in a macaroon as various caveats.
 */
public class MacaroonContext {

    private static final Logger LOGGER = LoggerFactory.getLogger(MacaroonContext.class);

    private FsPath root = FsPath.ROOT;
    private FsPath home = FsPath.ROOT;

    // FIXME: path is a relative path (relative to root) but uses FsPath, which
    // is intended to store only absolute paths in dCache namespace.
    private FsPath path = FsPath.ROOT;
    private String username;
    private OptionalLong maxUpload = OptionalLong.empty();
    private long uid = -1;
    private long[] gids;
    private EnumSet<Activity> activities = EnumSet.allOf(Activity.class);
    private String signature; // The macaroon's signature, if known.
    private Optional<Instant> expiry = Optional.empty();
    private String issueId;

    public void updateHome(String directory) throws InvalidCaveatException {
        LOGGER.debug("Updating home: {}", directory);
        home = root.resolve(directory);
    }

    public void setHome(FsPath path) {
        LOGGER.debug("Setting home to {}", path);
        home = path;
    }

    public Optional<FsPath> getHome() {
        return home == FsPath.ROOT ? Optional.empty() : Optional.of(home);
    }

    /**
     * Specify the new root value.  This does NOT modify the "path" path.
     */
    public void setRoot(FsPath newRoot) {
        LOGGER.debug("Setting root to {}", root);
        checkArgument(newRoot.hasPrefix(root), "Attempt to weaken root path");
        root = newRoot;
    }

    /**
     * Update the root value.  This will modify any non-/ "path" path.
     */
    public void updateRoot(String directory) throws InvalidCaveatException {
        FsPath newRoot = root.chroot(directory);

        FsPath delta = FsPath.ROOT.resolve(directory);

        FsPath pathCompare = delta;
        if (path.length() < delta.length()) {
            pathCompare = delta.drop(delta.length() - path.length());
        }

        checkCaveat(path.hasPrefix(pathCompare), "root extends outside of path");
        path = FsPath.ROOT.resolve(path.stripPrefix(pathCompare));

        if (home.hasPrefix(delta)) {
            home = FsPath.ROOT.resolve(home.stripPrefix(delta));
        } else {
            home = FsPath.ROOT;
        }

        root = newRoot;
    }

    public Optional<FsPath> getRoot() {
        return root == FsPath.ROOT ? Optional.empty() : Optional.of(root);
    }

    /**
     * Note that the supplied path is interpreted as being relative to the root.
     */
    public void setPath(String desiredPath) {
        path = FsPath.ROOT.chroot(desiredPath);
        LOGGER.debug("Setting path to {}", path);
    }

    public void updatePath(String directory) {
        path = path.chroot(directory);
    }

    /**
     * The returned value is a relative path that should be resolved against
     * the root path {@see getRoot()} to obtain the full path in dCache
     * namespace.
     */
    public Optional<String> getPath() {
        return path == FsPath.ROOT ? Optional.empty() : Optional.of(path.toString());
    }

    public void setUsername(String name) {
        checkArgument(!name.isEmpty(), "username is empty");
        checkArgument(username == null, "username is already set");
        username = name;
    }

    public String getUsername() {
        checkState(username != null, "username not specified");
        return username;
    }

    public void setUid(long id) {
        checkArgument(id >= 0, "invalid uid value");
        checkArgument(uid == -1, "uid is already set");
        uid = id;
    }

    public long getUid() {
        checkState(uid != -1, "uid not specified");
        return uid;
    }

    public void setGids(long[] id) {
        checkArgument(id.length > 0, "empty list of gids");
        checkArgument(gids == null, "gids are already set");
        gids = id;
    }

    public long[] getGids() {
        checkState(gids != null, "gid not specified");
        return gids;
    }

    public LongStream getGidStream() {
        return Arrays.stream(getGids());
    }

    public void updateAllowedActivities(EnumSet<Activity> newActivities)
          throws InvalidCaveatException {
        checkCaveat(activities.containsAll(newActivities), "attempt to enlarge activity set");
        activities = newActivities;
    }

    public void keepActivities(EnumSet<Activity> newActivities) {
        activities.retainAll(newActivities);
    }

    public void removeActivities(EnumSet<Activity> deniedActivities) {
        LOGGER.debug("Denying activities: {}", deniedActivities);
        activities.removeAll(deniedActivities);
    }

    public Optional<EnumSet<Activity>> getAllowedActivities() {
        /*  In general, any non-READ_METADATA activity only makes sense if
         *  the user is allowed to "enter a directory".  This requires the user
         *  isn't banned from the READ_METADATA activity, which would be the
         *  case if the "activity" caveat failed to include READ_METADATA
         *  activity.  Therefore, we treat the user's authorisation of any other
         *  activity as an implicit authorisation of the READ_METADATA activity.
         *
         *  Note that other caveats can impose additional Restrictions that
         *  limit READ_METADATA; e.g., the path caveat.  Implicitly authorising
         *  READ_METADATA in allowed activities does not stop those other
         *  Restrictions from being effective.
         */

        if (activities.size() == Activity.values().length ||
              activities.size() == Activity.values().length - 1 && !activities.contains(
                    Activity.READ_METADATA)) {
            // As an optimisation, treat an authorisation of all activities as
            // if the macaroon has no activity caveat.
            return Optional.empty();
        } else if (activities.isEmpty()) {
            return Optional.of(EnumSet.noneOf(Activity.class));
        } else {
            EnumSet<Activity> a = EnumSet.copyOf(activities);
            a.add(Activity.READ_METADATA);
            return Optional.of(a);
        }
    }

    public void updateExpiry(Instant newExpiry) {
        requireNonNull(newExpiry);

        if (!expiry.isPresent() || newExpiry.isBefore(expiry.get())) {
            expiry = Optional.of(newExpiry);
            LOGGER.debug("Updating expiry to {}", newExpiry);
        }
    }

    public Optional<Instant> getExpiry() {
        return expiry;
    }

    public void updateMaxUpload(long value) throws InvalidCaveatException {
        checkCaveat(value > 0, CaveatType.MAX_UPLOAD.getLabel() + " must be a positive value");
        long updatedValue = maxUpload.isPresent() ? Math.min(maxUpload.getAsLong(), value) : value;
        maxUpload = OptionalLong.of(updatedValue);
    }

    public OptionalLong getMaxUpload() {
        return maxUpload;
    }

    /**
     * Hexadecimal encoded macaroon signature.
     *
     * @param signature
     */
    public void setSignature(String signature) {
        this.signature = requireNonNull(signature);

    }

    public String getId() {
        checkState(signature != null, "Missing signature");
        return buildId(signature, issueId);
    }

    public static String buildId(String signature, String issueId) {
        // 6 bytes -> 48 bit: collision p=0.01 after 2.4x10^6 values, p=0.5 after 2x10^7.
        byte[] rawId = BaseEncoding.base16().lowerCase().decode(signature.substring(0, 12));
        String partialSignature = new String(Base64.getEncoder().encode(rawId),
              StandardCharsets.US_ASCII);

        return issueId == null ? partialSignature : issueId + ":" + partialSignature;
    }

    public void updateIssueId(String value) {
        issueId = requireNonNull(value);
    }

    public String getIssueId() {
        return issueId;
    }
}
