package org.dcache.pool.movers;

import static com.google.common.base.Preconditions.checkState;
import static com.google.common.collect.Lists.newArrayList;
import static org.dcache.util.ByteUnit.KiB;
import static org.dcache.util.Exceptions.messageOrClassName;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.DiscreteDomain;
import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import com.google.common.collect.TreeRangeSet;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.nio.ByteBuffer;
import java.nio.channels.ReadableByteChannel;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import javax.annotation.concurrent.GuardedBy;
import org.dcache.pool.repository.ForwardingRepositoryChannel;
import org.dcache.pool.repository.RepositoryChannel;
import org.dcache.util.Checksum;
import org.dcache.util.ChecksumType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A wrapper for RepositoryChannel that computes a digest on the fly during write as long as all
 * writes are sequential.
 */
public class ChecksumChannel extends ForwardingRepositoryChannel {

    private static final Logger LOGGER =
          LoggerFactory.getLogger(ChecksumChannel.class);

    /**
     * Inner channel to which all operations are delegated.
     */
    @VisibleForTesting
    RepositoryChannel _channel;

    /**
     * Digest used for computing the checksum during write.
     */
    private final List<MessageDigest> _digests;

    /**
     * Cached checksum after getChecksums is called the first time.
     */
    @GuardedBy("this")
    private Set<Checksum> _finalChecksums;

    /**
     * RangeSet to keep track of written bytes
     */
    private final RangeSet<Long> _dataRangeSet = TreeRangeSet.create();

    /**
     * The offset where the checksum was calculated.
     */
    @GuardedBy("_digests")
    private long _nextChecksumOffset = 0L;

    /**
     * Flag to indicate whether it is still possible to calculated a checksum
     */
    private volatile boolean _isChecksumViable = true;

    /**
     * Buffer to be used for reading data back from the inner channel for checksum calculations.
     */
    @VisibleForTesting
    @GuardedBy("_digests")
    ByteBuffer _readBackBuffer = ByteBuffer.allocate(KiB.toBytes(256));

    /*
     * Static buffer with zeros shared with in all instances of ChecksumChannel.
     */
    private static final ByteBuffer ZERO_BUFFER = ByteBuffer
          .allocate(KiB.toBytes(256))
          .asReadOnlyBuffer();

    /**
     * Buffer to be used for feeding the checksum digester with 0s to fill up gaps in ranges.
     */
    @VisibleForTesting
    ByteBuffer _zerosBuffer = ZERO_BUFFER.duplicate();

    public ChecksumChannel(RepositoryChannel inner, Set<ChecksumType> types) {
        _channel = inner;
        _digests = types.stream()
              .map(t -> t.createMessageDigest())
              .collect(Collectors.toList());
    }

    /**
     * Ensure that a Checksum is calculated for the supplied ChecksumType.  If the ChecksumType is
     * already registered then this method does nothing, otherwise the ChecksumChannel is updated to
     * calculate the new ChecksumType. If the ChecksumChannel has accepted a contiguous range of
     * data from offset 0 then this method will reread that contiguous range.
     *
     * @param type The algorithm this ChecksumChannel should calculate.
     * @throws IOException if the Channel has already started accepting data and an attempt to
     *                     reread data from disk fails.
     */
    public void addType(ChecksumType type) throws IOException {
        synchronized (_digests) {
            if (_digests.stream()
                  .map(MessageDigest::getAlgorithm)
                  .noneMatch(t -> t.equals(type.getName()))) {
                MessageDigest digest = type.createMessageDigest();

                if (_isChecksumViable) {
                    try {
                        updateFromChannel(Collections.singleton(digest), 0L, _nextChecksumOffset);
                    } catch (IOException e) {
                        throw new IOException("Failed when reading received data: "
                              + messageOrClassName(e), e);
                    }
                }

                _digests.add(digest);
            }
        }
    }

    @Override
    protected RepositoryChannel delegate() {
        return _channel;
    }

    @Override
    public int write(ByteBuffer buffer, long position) throws IOException {
        int bytes;
        if (_isChecksumViable) {
            ByteBuffer readOnly = buffer.asReadOnlyBuffer();
            bytes = _channel.write(buffer, position);
            updateChecksum(readOnly, position, bytes);
        } else {
            bytes = _channel.write(buffer, position);
        }
        return bytes;
    }

    @Override
    public long transferFrom(ReadableByteChannel src, long position,
          long count) throws IOException {
        _isChecksumViable = false;
        return _channel.transferFrom(src, position, count);
    }

    @Override
    public int write(ByteBuffer src) throws IOException {
        int bytes;
        if (_isChecksumViable) {
            bytes = writeWithChecksumUpdate(src);
        } else {
            bytes = _channel.write(src);
        }
        return bytes;
    }

    @Override
    public long write(ByteBuffer[] srcs, int offset, int length)
          throws IOException {
        long bytes = 0;
        if (_isChecksumViable) {
            for (int i = offset; i < offset + length; i++) {
                bytes += writeWithChecksumUpdate(srcs[i]);
            }
        } else {
            bytes = _channel.write(srcs, offset, length);
        }
        return bytes;
    }

    @Override
    public synchronized long write(ByteBuffer[] srcs) throws IOException {
        return write(srcs, 0, srcs.length);
    }

    @Override
    public synchronized void close() throws IOException {
        if (_finalChecksums == null) {
            _finalChecksums = finalizeChecksums();
        }

        super.close();
    }

    /**
     * Provide the checksums calculated from writing data to this channel.
     * Calling this method is only valid after the channel has been closed.
     * @return on-the-fly checksums of this channel.
     */
    public synchronized Set<Checksum> getChecksums() {
        checkState(!isOpen(), "Checksums are only available after the channel is closed.");
        return _finalChecksums;
    }

    /**
     * Complete checksum calculation.  For non-sparse files, this method is
     * trivial.  For sparse files, any gaps are assumed to be zero bytes and the
     * checksum is calculated accordingly.  If there is any problem then this
     * method returns an empty set.
     * @return Checksums for this channel.
     */
    private Set<Checksum> finalizeChecksums() {

        if (!_isChecksumViable) {
            return Collections.emptySet();
        }

        // we need to synchronize on rangeSet and digest get exclusive access
        synchronized (_dataRangeSet) {
            synchronized (_digests) {
                try {
                    if (_dataRangeSet.asRanges().size() > 1
                            || (_dataRangeSet.asRanges().size() == 1 && _nextChecksumOffset == 0)) {
                        feedZerosToDigesterForRangeGaps();
                    }

                    return _digests.stream()
                          .map(Checksum::new)
                          .collect(Collectors.toSet());
                } catch (IOException e) {
                    LOGGER.info("Unable to generate checksum of sparse file: {}", e.toString());
                    return Collections.emptySet();
                }
            }
        }
    }

    private void feedZerosToDigesterForRangeGaps() throws IOException {
        ArrayList<Range<Long>> complement = newArrayList(
              _dataRangeSet.complement().subRangeSet(Range.closed(0L, size())).asRanges());
        complement.sort((r1, r2) -> r1.lowerEndpoint().compareTo(r2.lowerEndpoint()));

        for (Range<Long> range : complement) {

            long bytesToWrite = range.upperEndpoint() - range.lowerEndpoint();
            long chunkOffset = range.lowerEndpoint();

            while (bytesToWrite > 0) {
                _zerosBuffer.clear();
                long chunkSize = Math.min(_zerosBuffer.capacity(), bytesToWrite);
                _zerosBuffer.limit((int) chunkSize);

                updateChecksum(_zerosBuffer, chunkOffset, _zerosBuffer.limit());

                chunkOffset += chunkSize;
                bytesToWrite -= chunkSize;
            }
        }
    }

    private int writeWithChecksumUpdate(ByteBuffer src) throws IOException {
        int writtenBytes;
        ByteBuffer readOnly = src.asReadOnlyBuffer();
        long updatePosition = position();
        writtenBytes = _channel.write(src);
        updateChecksum(readOnly, updatePosition, writtenBytes);

        return writtenBytes;
    }

    /**
     * @param buffer   buffer containing the data
     * @param position position of the data in the target file
     * @param bytes    number of bytes to use from the input data
     * @throws IOException
     */
    @VisibleForTesting
    void updateChecksum(ByteBuffer buffer, long position, int bytes) throws IOException {

        if (bytes == 0) {
            return;
        }

        if (bytes < buffer.remaining()) {
            buffer.limit(buffer.position() + bytes);
        }

        Range<Long> writeRange = Range.closed(position, position + buffer.remaining() - 1)
              .canonical(DiscreteDomain.longs());
        Range<Long> fileStartRange;

        synchronized (_dataRangeSet) {

            RangeSet<Long> overlappingRanges = _dataRangeSet.subRangeSet(writeRange);
            if (!overlappingRanges.isEmpty()) {
                _isChecksumViable = false;
                LOGGER.info("On-transfer checksum aborted due to overlapping writes from client.");
                return;
            }

            fileStartRange = _dataRangeSet.rangeContaining(0L);
            boolean canCalculateChecksum = position == 0 || (fileStartRange != null
                  && fileStartRange.upperEndpoint() == position);

            _dataRangeSet.add(writeRange);
            if (!canCalculateChecksum) {
                return;
            }

            // get it again as we may have merged two segments
            fileStartRange = _dataRangeSet.rangeContaining(0L);
        }

        synchronized (_digests) {
            /*
             * we are one of the threads which got the merge into continues block.
             * Nevertheless, there may be a different thread which needs to update
             * ahead of us. Wait for our turn.
             */
            while (_nextChecksumOffset != position) {
                try {
                    _digests.wait();
                } catch (InterruptedException e) {
                    throw new InterruptedIOException();
                }
            }

            long bytesToRead = fileStartRange.upperEndpoint() - position;

            // update current buffer and then keep procesing following blocks, if any
            bytesToRead -= buffer.remaining();
            // update offset prior digest calculation as digests#update will update position in the buffer
            _nextChecksumOffset += buffer.remaining();

            _digests.forEach(d -> d.update(buffer.duplicate()));

            long expectedOffsetAfterRead = _nextChecksumOffset + bytesToRead;
            try {
                updateFromChannel(_digests, _nextChecksumOffset, bytesToRead);
            } finally {
                _nextChecksumOffset = expectedOffsetAfterRead;
                _digests.notifyAll();
            }
        }
    }

    @GuardedBy("_digests")
    private void updateFromChannel(Collection<MessageDigest> digests, long offset, long bytesToRead)
          throws IOException {
        try {
            while (bytesToRead > 0) {
                _readBackBuffer.clear();
                _readBackBuffer.limit((int) Math.min(_readBackBuffer.capacity(), bytesToRead));
                int bytesRead = _channel.read(_readBackBuffer, offset);

                if (bytesRead < 0) {
                    throw new IOException(
                          "Checksum: Unexpectedly hit end-of-stream while reading data back from channel.");
                }

                _readBackBuffer.flip();

                digests.forEach(d -> d.update(_readBackBuffer.asReadOnlyBuffer()));

                bytesToRead -= bytesRead;
                offset += bytesRead;
            }
        } catch (IOException | RuntimeException e) {
            _isChecksumViable = false;
            throw e;
        }
    }
}
